package com.jetbyte.cache4;

import java.util.Map;
import java.util.HashMap;
import java.util.LinkedList;

public class TimeLimitedCache extends CacheImpl
{
   public TimeLimitedCache(String name)
   {
      super(name, TimeLimitedCache.class.getName());
   }

   protected Object findObject(Cache.Key key)
   {      
      Cache.ObjectRef ref = (Cache.ObjectRef)cache.get(key);

      Object obj = null;

      if (ref != null)
      {
         obj = ref.getObject();

         lruList.remove(ref);
         lruList.addLast(ref);            
      }

      return obj;
   }

   protected void addObject(Cache.Key key, Object obj)
   {
      Cache.ObjectRef ref = new Cache.ObjectRef(this, key, obj); 

      cache.put(key, ref);

      lruList.addLast(ref); 
   }

   protected void removeObject(Cache.Key key)
   {
      cache.remove(key);
   }

   synchronized public int entries()
   {
      return cache.size();
   }


   protected void flushCache()
   {
      cache.clear();

      lruList.clear();
   }

   protected long getReapPeriod()
   {
      return scanEveryMillis;
   }

   synchronized private void reap()
   {
      boolean done = (lruList.size() == 0);
      
      System.out.println("Reaping - " + name());
      System.out.println("LRU size = " + lruList.size());

      while (!done)
      {
         Cache.ObjectRef ref = (Cache.ObjectRef)lruList.getFirst();

         if (timedOut(ref))
         {
            System.out.println("Reference reaped: - " + ref.getKey());

            ref.removeFromCache();

            lruList.removeFirst();

            done = (lruList.size() == 0);
         }
         else
         {
            done = true;
         }
      }
   }

   private boolean timedOut(Cache.ObjectRef ref) 
   {
      return (System.currentTimeMillis() - ref.getLastUsedTime() > entryTimeoutMillis);
   }


   private long entryTimeoutMillis = 30000;  // 30 secs
   
   private long scanEveryMillis = 10000; // 10 secs

   private Map cache = new HashMap(89);  
   
   private LinkedList lruList = new LinkedList(); 

   private ReaperThread reaper = ReaperThread.create(this);

   private static class ReaperThread extends Thread
   {
      private ReaperThread(TimeLimitedCache cache)
      {
         this.cache = cache;      
      }

      public void run()
      {
         // Spread out the reaping for multiple caches with
         // the same reap period by waiting for a random
         // portion of the reap period before starting up...

         pause((long)(Math.random() * cache.getReapPeriod()));

         while (true)
         {
            pause(cache.getReapPeriod());
            
            cache.reap();
         }            
      }

      private void pause(long timeout)
      {
         try
         {
            synchronized(this)
            {
               wait(timeout);
            }
         }
         catch(InterruptedException e)
         {

         }
      }

      private static ReaperThread create(TimeLimitedCache cache)
      {
         ReaperThread reaper = new ReaperThread(cache);

         reaper.setDaemon(true);

         reaper.start();
         
         return reaper;
      }

      private TimeLimitedCache cache;
   }
}
